import logging
import itertools
import numpy as np
import pandas as pd
from pymoo.factory import get_performance_indicator
from pyaugmecon.logs import Logs
from pyaugmecon.model import Model
from pyaugmecon.helper import Helper, Timer
from pyaugmecon.options import Options
from pyaugmecon.queue_handler import QueueHandler
from pyaugmecon.process_handler import ProcessHandler


class PyAugmecon(object):
    def __init__(self, model, opts, solver_opts={}):
        self.opts = Options(opts, solver_opts)
        self.logs = Logs(self.opts)
        self.logger = logging.getLogger(self.opts.log_name)
        self.logger.setLevel(logging.INFO)
        self.opts.log()
        self.model = Model(model, self.opts)
        self.opts.check(self.model.n_obj)

    def find_solutions(self):
        self.model.progress.set_message("finding solutions")

        if self.model.min_obj:
            grid_range = list(reversed(range(self.opts.gp)))
        else:
            grid_range = range(self.opts.gp)

        indices = [tuple([n for n in grid_range]) for _ in self.model.iter_obj2]
        self.cp = list(itertools.product(*indices))
        self.cp = [i[::-1] for i in self.cp]

        self.model.pickle()
        self.queues = QueueHandler(self.cp, self.opts)
        self.queues.split_work()
        self.procs = ProcessHandler(self.opts, self.model, self.queues)

        self.procs.start()
        self.sols = self.queues.get_result()
        self.procs.join()
        self.model.clean()

    def process_solutions(self):
        def keep_undominated(pts, min):
            pts = np.array(pts)
            undominated = np.ones(pts.shape[0], dtype=bool)
            for i, c in enumerate(pts):
                if undominated[i]:
                    if min:
                        undominated[undominated] = np.any(pts[undominated] < c, axis=1)
                    else:
                        undominated[undominated] = np.any(pts[undominated] > c, axis=1)
                    undominated[i] = True

            return pts[undominated, :]

        # Remove duplicate solutions
        self.sols = list(set(tuple(self.sols)))
        self.num_sols = len(self.sols)

        # Remove duplicate solutions due to numerical issues by rounding
        self.unique_sols = [
            tuple(round(sol, self.opts.round) for sol in item) for item in self.sols
        ]
        self.unique_sols = list(set(tuple(self.unique_sols)))
        self.num_unique_sols = len(self.unique_sols)

        # Remove dominated solutions
        self.unique_pareto_sols = keep_undominated(self.unique_sols, self.model.min_obj)
        self.num_unique_pareto_sols = len(self.unique_pareto_sols)

    def output_excel(self):
        writer = pd.ExcelWriter(f"{self.logs.logdir}{self.opts.log_name}.xlsx")
        pd.DataFrame(self.model.e).to_excel(writer, "e_points")
        pd.DataFrame(self.model.payoff).to_excel(writer, "payoff_table")
        pd.DataFrame(self.sols).to_excel(writer, "sols")
        pd.DataFrame(self.unique_sols).to_excel(writer, "unique_sols")
        pd.DataFrame(self.unique_pareto_sols).to_excel(writer, "unique_pareto_sols")
        writer.save()

    def get_hv_indicator(self):
        ref = np.diag(self.model.payoff)
        hv = get_performance_indicator("hv", ref_point=ref)
        self.hv_indicator = hv.do(self.unique_pareto_sols)

    def solve(self):
        self.runtime = Timer()
        self.model.construct_payoff()
        self.model.find_obj_range()
        self.model.convert_prob()
        self.find_solutions()
        self.process_solutions()
        self.get_hv_indicator()
        if self.opts.output_excel:
            self.output_excel()

        self.runtime = round(self.runtime.get(), 2)
        Helper.clear_line()
        print(
            f"Solved {self.model.models_solved.value()} models for "
            f"{self.num_unique_pareto_sols} unique Pareto solutions in "
            f"{self.runtime} seconds"
        )

        self.logger.info(Helper.separator())
        self.logger.info(f"Runtime: {self.runtime} seconds")
        self.logger.info(f"Models solved: {self.model.models_solved.value()}")
        self.logger.info(f"Infeasibilities: {self.model.infeasibilities.value()}")
        self.logger.info(f"Solutions: {self.num_sols}")
        self.logger.info(f"Unique solutions: {self.num_unique_sols}")
        self.logger.info(f"Unique Pareto solutions: {self.num_unique_pareto_sols}")
        self.logger.info(f"Hypervolume indicator: {self.hv_indicator}")



import queue
import logging
import numpy as np
from multiprocessing import Queue
from pyaugmecon.options import Options


class QueueHandler(object):
    def __init__(self, work, opts: Options):
        self.work = work
        self.opts = opts
        self.logger = logging.getLogger(opts.log_name)
        self.result_q = Queue()

    def get_longest_q(self):
        q_length = [q.qsize() for q in self.job_qs]
        if all(q == 0 for q in q_length):
            return None
        else:
            return q_length.index(max(q_length))

    def get_work(self, i):
        try:
            return self.job_qs[i].get_nowait()
        except queue.Empty:
            if self.opts.redivide_work and self.get_longest_q():
                return self.get_work(self.get_longest_q())
            else:
                if self.opts.process_logging:
                    self.logger.info(f"PID: {i} exited")

                self.result_q.put("STOP")
                return None

    def put_result(self, result):
        self.result_q.put(result)

    def get_result(self):
        results = []

        for _ in range(self.proc_count):
            while True:
                result = self.result_q.get()
                if result == "STOP":
                    break
                else:
                    results.append(result)

        return results

    def empty_job_qs(self):
        for i in range(self.proc_count):
            while True:
                try:
                    self.job_qs[i].get_nowait()
                except queue.Empty:
                    break

    def split_work(self):
        blocks = [
            self.work[i : i + self.opts.gp]
            for i in range(0, len(self.work), self.opts.gp)
        ]
        blocks = np.array_split(np.array(blocks), self.opts.cpu_count)
        blocks = [x for x in blocks if x.size > 0]  # Remove empty blocks
        self.proc_count = len(blocks)

        self.logger.info(f"Dividing grid over {self.proc_count} process(es)")
        self.job_qs = [Queue() for _ in range(self.proc_count)]

        for i, b in enumerate(blocks):
            for item in b:
                item = [tuple(x) for x in item.tolist()]
                self.job_qs[i].put_nowait(item)



import os
import logging
import cloudpickle
import numpy as np
import pyomo.environ as pyo
from pyaugmecon.options import Options
from pyaugmecon.helper import Counter, ProgressBar
from pyomo.core.base import (
    Var,
    ConstraintList,
    maximize,
    minimize,
    Set,
    Param,
    NonNegativeReals,
    Any,
)


class Model(object):
    def __init__(self, model: pyo.ConcreteModel, opts: Options):
        self.model = model
        self.opts = opts
        self.logger = logging.getLogger(opts.log_name)

        self.n_obj = len(self.model.obj_list)
        self.iter_obj = range(self.n_obj)
        self.iter_obj2 = range(self.n_obj - 1)
        self.min_obj = self.obj_sense(0) == minimize

        # Setup progress bar
        self.to_solve = self.opts.gp ** (self.n_obj - 1) + self.n_obj ** 2
        self.progress = ProgressBar(Counter(), self.to_solve)

        self.models_solved = Counter()
        self.infeasibilities = Counter()

        if self.n_obj < 2:
            raise Exception("Too few objective functions provided")

    def obj(self, i):
        return self.model.obj_list[i + 1]

    def obj_val(self, i):
        return self.obj(i)()

    def obj_expr(self, i):
        return self.obj(i).expr

    def obj_sense(self, i):
        return self.obj(i).sense

    def slack_val(self, i):
        return self.model.Slack[i + 1].value

    def obj_activate(self, i):
        self.obj(i).activate()

    def obj_deactivate(self, i):
        self.obj(i).deactivate()

    def solve(self):
        opt = pyo.SolverFactory(self.opts.solver_name, solver_io=self.opts.solver_io)
        opt.options.update(self.opts.solver_opts)
        self.result = opt.solve(self.model)
        self.term = self.result.solver.termination_condition
        self.status = self.result.solver.status

    def pickle(self):
        model_file = open(self.opts.model_fn, "wb")
        cloudpickle.dump(self.model, model_file)
        del self.model

    def unpickle(self):
        model_file = open(self.opts.model_fn, "rb")
        self.model = cloudpickle.load(model_file)

    def clean(self):
        if os.path.exists(self.opts.model_fn):
            os.remove(self.opts.model_fn)

    def is_optimal(self):
        return (
            self.status == pyo.SolverStatus.ok
            and self.term == pyo.TerminationCondition.optimal
        )

    def is_infeasible(self):
        return (
            self.term == pyo.TerminationCondition.infeasible
            or self.term == pyo.TerminationCondition.infeasibleOrUnbounded
        )

    def construct_payoff(self):
        self.logger.info("Constructing payoff")
        self.progress.set_message("constructing payoff")

        def set_payoff(i, j):
            self.obj_activate(j)
            self.solve()
            self.progress.increment()
            self.payoff[i, j] = self.obj_val(j)
            self.obj_deactivate(j)

        self.payoff = np.full((self.n_obj, self.n_obj), np.inf)
        self.model.pcon_list = ConstraintList()

        # Independently optimize each objective function (diagonal elements)
        for i in self.iter_obj:
            for j in self.iter_obj:
                if i == j:
                    set_payoff(i, j)

        # Optimize j having all the i as constraints (off-diagonal elements)
        for i in self.iter_obj:
            self.model.pcon_list.add(expr=self.obj_expr(i) == self.payoff[i, i])

            for j in self.iter_obj:
                if i != j:
                    set_payoff(i, j)
                    self.model.pcon_list.add(expr=self.obj_expr(j) == self.payoff[i, j])

            self.model.pcon_list.clear()

    def find_obj_range(self):
        self.logger.info("Finding objective function range")

        # Gridpoints of p-1 objective functions that are used as constraints
        self.e = np.zeros((self.n_obj - 1, self.opts.gp))
        self.obj_range = np.zeros(self.n_obj - 1)

        for i in self.iter_obj2:
            if self.opts.nadir_p:
                min = self.opts.nadir_p[i]
            else:
                min = self.opts.nadir_r * np.min(self.payoff[:, i + 1], 0)

            max = np.max(self.payoff[:, i + 1], 0)
            self.obj_range[i] = max - min
            self.e[i] = [
                min + j * (self.obj_range[i] / (self.opts.gp - 1))
                for j in range(0, self.opts.gp)
            ]

    def convert_prob(self):
        self.logger.info("Converting optimization problem")

        self.model.con_list = ConstraintList()

        # Set of objective functions
        self.model.Os = Set(ordered=True, initialize=[o + 2 for o in self.iter_obj2])

        # Slack for objectives introduced as constraints
        self.model.Slack = Var(self.model.Os, within=NonNegativeReals)
        self.model.e = Param(
            self.model.Os,
            initialize=[np.nan for _ in self.model.Os],
            within=Any,
            mutable=True,
        )  # RHS of constraints

        # Add p-1 objective functions as constraints
        for o in range(1, self.n_obj):
            self.model.obj_list[1].expr += self.opts.eps * (
                10 ** (-1 * (o - 1)) * self.model.Slack[o + 1] / self.obj_range[o - 1]
            )

            if self.model.obj_list[o + 1].sense == minimize:
                self.model.con_list.add(
                    expr=self.model.obj_list[o + 1].expr + self.model.Slack[o + 1]
                    == self.model.e[o + 1]
                )
            elif self.model.obj_list[o + 1].sense == maximize:
                self.model.con_list.add(
                    expr=self.model.obj_list[o + 1].expr - self.model.Slack[o + 1]
                    == self.model.e[o + 1]
                )



import time
import logging
from multiprocessing import cpu_count
from pyaugmecon.helper import Helper


class Options(object):
    def __init__(self, opts: dict, solver_opts: dict):
        self.name = opts.get("name", "Undefined")
        self.gp = opts.get("grid_points")
        self.nadir_p = opts.get("nadir_points")
        self.eps = opts.get("penalty_weight", 1e-3)
        self.round = opts.get("round_decimals", 9)
        self.nadir_r = opts.get("nadir_ratio", 1)
        self.logdir = opts.get("logging_folder", "logs")
        self.early_exit = opts.get("early_exit", True)
        self.bypass = opts.get("bypass_coefficient", True)
        self.flag = opts.get("flag_array", True)
        self.cpu_count = opts.get("cpu_count", cpu_count())
        self.redivide_work = opts.get("redivide_work", True)
        self.model_fn = opts.get("pickle_file", "model.p")
        self.shared_flag = opts.get("shared_flag", True)
        self.solver_name = opts.get("solver_name", "gurobi")
        self.solver_io = opts.get("solver_io", "python")
        self.output_excel = opts.get("output_excel", True)
        self.process_logging = opts.get("process_logging", False)
        self.process_timeout = opts.get("process_timeout", None)

        self.solver_opts = solver_opts
        self.solver_opts["MIPGap"] = solver_opts.get("MIPGap", 0.0)
        self.solver_opts["NonConvex"] = solver_opts.get("NonConvex", 2)

        self.time_created = time.strftime("%Y%m%d-%H%M%S")
        self.log_name = self.name + "_" + str(self.time_created)

    def log(self):
        self.logger = logging.getLogger(self.log_name)
        self.logger.info(f"Name: {self.name}")
        self.logger.info(f"Grid points: {self.gp}")
        self.logger.info(f"Nadir points: {self.nadir_p}")
        self.logger.info(f"Penalty weight: {self.eps}")
        self.logger.info(f"Early exit: {self.early_exit}")
        self.logger.info(f"Bypass coefficient: {self.bypass}")
        self.logger.info(f"Flag array: {self.flag}")
        self.logger.info(f"CPU Count: {self.cpu_count}")
        self.logger.info(f"Redivide work: {self.redivide_work}")
        self.logger.info(f"Shared flag array: {self.shared_flag}")
        self.logger.info(Helper.separator())

    def check(self, num_objfun):
        if not self.gp:
            raise Exception("No number of grid points provided")

        if self.nadir_p and len(self.nadir_p) != num_objfun - 1:
            raise Exception("Too many or too few nadir points provided")



import time
from multiprocessing import Value, Lock


class Helper(object):
    def clear_line():
        print(" " * 80, end="\r", flush=True)

    def separator():
        return "=" * 30


class Counter(object):
    def __init__(self, init_val: int = 0):
        self.val = Value("i", init_val)
        self.lock = Lock()

    def increment(self):
        with self.lock:
            self.val.value += 1

    def value(self):
        with self.lock:
            return self.val.value


class Timer(object):
    def __init__(self):
        self.start_time = time.time()

    def get(self):
        return time.time() - self.start_time


class ProgressBar(object):
    def __init__(self, counter: Counter, total: int, init_message: str = ""):
        self.counter = counter
        self.total = total
        self.message = init_message
        self.bar = ""

    def set_message(self, message):
        self.message = message
        Helper.clear_line()
        self.print(True)

    def print(self, force):
        bar_len = 40

        progress = self.counter.value() / float(self.total)
        filled_len = int(round(bar_len * progress, 1))
        percents = round(100.0 * progress, 1)
        bar = "=" * filled_len + "-" * (bar_len - filled_len)

        if self.bar != bar or force:
            self.bar = bar
            print(f"[{bar}] {percents}% ... ({self.message})", end="\r", flush=True)

    def increment(self):
        self.counter.increment()
        self.print(False)



import logging
from multiprocessing import Process
from pyaugmecon.flag import Flag
from pyaugmecon.model import Model
from pyaugmecon.options import Options
from pyaugmecon.queue_handler import QueueHandler


class SolverProcess(Process):
    def __init__(
        self, p_num, opts: Options, model: Model, queues: QueueHandler, flag: Flag
    ):
        Process.__init__(self)
        self.p_num = p_num
        self.opts = opts
        self.model = model
        self.queues = queues
        self.flag = flag

    def run(self):
        jump = 0
        if self.opts.process_logging:
            logger = logging.getLogger(self.opts.log_name)
            logger.setLevel(logging.INFO)

        self.model.unpickle()

        while True:
            work = self.queues.get_work(self.p_num)

            if work:
                for c in work:
                    log = f"Process: {self.p_num}, index: {c}, "

                    cp_start = self.opts.gp - 1 if self.model.min_obj else 0
                    cp_end = 0 if self.model.min_obj else self.opts.gp - 1

                    self.model.progress.increment()

                    def do_jump(i, jump):
                        return min(jump, abs(cp_end - i))

                    def bypass_range(i):
                        if i == 0:
                            return range(c[i], c[i] + 1)
                        elif self.model.min_obj:
                            return range(c[i] - b[i], c[i] + 1)
                        else:
                            return range(c[i], c[i] + b[i] + 1)

                    def early_exit_range(i):
                        if i == 0:
                            return range(c[i], c[i] + 1)
                        elif self.model.min_obj:
                            return range(c[i], cp_start)
                        else:
                            return range(c[i], cp_end)

                    if self.opts.flag and self.flag.get(c) != 0 and jump == 0:
                        jump = do_jump(c[0] - 1, self.flag.get(c))

                    if jump > 0:
                        jump = jump - 1
                        continue

                    for o in self.model.iter_obj2:
                        log += f"e{o + 1}: {self.model.e[o, c[o]]}, "
                        self.model.model.e[o + 2] = self.model.e[o, c[o]]

                    self.model.obj_activate(0)
                    self.model.solve()
                    self.model.models_solved.increment()

                    if self.opts.early_exit and self.model.is_infeasible():
                        self.model.infeasibilities.increment()
                        if self.opts.flag:
                            self.flag.set(
                                early_exit_range, self.opts.gp, self.model.iter_obj2
                            )
                        jump = do_jump(c[0], self.opts.gp)

                        log += "infeasible"
                        if self.opts.process_logging:
                            logger.info(log)
                        continue
                    elif self.opts.bypass and self.model.is_optimal():
                        b = []

                        for i in self.model.iter_obj2:
                            step = self.model.obj_range[i] / (self.opts.gp - 1)
                            slack = round(self.model.slack_val(i + 1))
                            b.append(int(slack / step))

                        log += f"jump: {b[0]}, "

                        if self.opts.flag:
                            self.flag.set(bypass_range, b[0] + 1, self.model.iter_obj2)
                        jump = do_jump(c[0], b[0])

                    sols = []

                    sols.append(
                        self.model.obj_val(0)
                        - self.opts.eps
                        * sum(
                            10 ** (-1 * (o))
                            * self.model.slack_val(o + 1)
                            / self.model.obj_range[o]
                            for o in self.model.iter_obj2
                        )
                    )

                    for o in self.model.iter_obj2:
                        sols.append(self.model.obj_val(o + 1))

                    self.queues.put_result(tuple(sols))

                    log += f"solutions: {sols}"
                    if self.opts.process_logging:
                        logger.info(log)
            else:
                break
